#!/usr/bin/env python
# coding: utf-8


#č nazvy proměnných jsou v angličtině
#č Ale komenty teda ne)

import numpy as np
from matplotlib import colors
import matplotlib.tri as mtri

# copied from simplex module
# just don't want every time import simplex and its dependencies
def get_events(sx, simplices=None):
    """
    Metoda musí simplexům přiřazovat jev 
    0=success, 1=failure, 2=mix
    """
    if simplices is None:
        simplices = sx.tri.simplices
    
    in_failure = np.isin(simplices, sx.sample_box.failure_points)
    has_failure = in_failure.any(axis=1)
    all_failure = in_failure.all(axis=1)
    return np.int8(np.where(has_failure, np.where(all_failure, 1, 2), 0))


def get_g_model_wireframe_data(shape_share, limits, space='R', ngrid=50):
    xmin, xmax, ymin, ymax = limits
    
    # endpoint=True by default
    # we'll get len(.) == ngrid
    # ngrid 51 is an default maximum for MPL
    # (otherwise MPL will downsample)
    x = np.linspace(xmin, xmax, ngrid)
    y = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(x, y)
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    
    if space == 'R':
        g_values = shape_share.gm(XY).g_values
    else:
        XY_sample = shape_share.f_model.new_sample(XY, space)
        g_values = shape_share.gm(XY_sample).g_values
    
    Z = g_values.reshape(ngrid, ngrid)
    return X, Y, Z


#č napadlo mě, že bych mohl matplotlibovskému Axes
#č přiřazovat (připsavat, zadávat) atribut 'space'
#č Daválo by to smysl, ne? U všeho ostatního, u sample boksů
#č nejsem jist, ale mám pocit, že na jedném sablotu někdo potřebuje
#č kreslit z různejch prostoru.
#č Zkrátka, funkce v tomto modulu požadujou aby 
#č ax.space a ax.sample_box byl nastaven!

# ax.space and ax.sample_box attributes should (expected to) be set up!


def scatter_points(ax3d, zs=None, markers=('P', 'X', 'h', 'H'), **kwargs):
    sample = getattr(ax3d.sample_box, ax3d.space)
    nsim = len(sample)
    if zs is not None:
        x, y = sample[:,:2].T
    elif ax3d.sample_box.nvar == 2:
        x, y = sample[:,:2].T
        #zs = ax3d.sample_box.pdf(ax3d.space)
        zs = ax3d.sample_box.g_values
    else:
        x, y, zs = sample[:,:3].T
    
    failsi = ax3d.sample_box.failsi
    
    try: # proxy denotes to implicitly-known values
        proxy = ax3d.sample_box.proxy
    except AttributeError:
        proxy = np.full(nsim, False, dtype=bool)
    
    sm, fm, psm, pfm = markers
    
    mask = np.all((~failsi, ~proxy), axis=0)
    success = ax3d.scatter(x[mask], y[mask], zs=zs[mask], c='g', marker=sm, **kwargs)
    
    mask = np.all((failsi, ~proxy), axis=0)
    failures = ax3d.scatter(x[mask], y[mask], zs=zs[mask], c='r', marker=fm, **kwargs)
    
    mask = np.all((~failsi, proxy), axis=0)
    proxy_successes = ax3d.scatter(x[mask], y[mask], zs=zs[mask],\
                             c='#77AC30', marker=psm, **kwargs)
    
    mask = np.all((failsi, proxy), axis=0)
    proxy_failures = ax3d.scatter(x[mask], y[mask], zs=zs[mask],\
                             c='#D95319', marker=pfm, **kwargs)
    
    return success, failures, proxy_successes, proxy_failures



def plot_points(ax3d, zs=None, markers=('P', 'X', 'h', 'H'), ls='', **kwargs):
    sample = getattr(ax3d.sample_box, ax3d.space)
    nsim = len(sample)
    if zs is not None:
        x, y = sample[:,:2].T
    elif ax3d.sample_box.nvar == 2:
        x, y = sample[:,:2].T
        #zs = ax3d.sample_box.pdf(ax3d.space)
        zs = ax3d.sample_box.g_values
    else:
        x, y, zs = sample[:,:3].T
    
    failsi = ax3d.sample_box.failsi
    
    try: # proxy denotes to implicitly-known values
        proxy = ax3d.sample_box.proxy
    except AttributeError:
        proxy = np.full(nsim, False, dtype=bool)
    
    sm, fm, psm, pfm = markers
    
    plot_list = []
    
    #č byl jsem svědkem, že matplotlib zlobil ve 3D 
    #č kvůli tomu, že nebyl žádný safe vzorek
    #č proto raději budu přidávat tečky podmíněne
    mask = np.all((~failsi, ~proxy), axis=0)
    if np.any(mask): #success
        plot_list.append(ax3d.plot(x[mask], y[mask], mec='g', mfc='g',\
                         marker=sm, ls=ls, **kwargs))
    
    mask = np.all((failsi, ~proxy), axis=0)
    if np.any(mask): #failures
        plot_list.append(ax3d.plot(x[mask], y[mask], mec='r', mfc='r',\
                         marker=fm, ls=ls, **kwargs))
    
    mask = np.all((~failsi, proxy), axis=0)
    if np.any(mask): #proxy_successes
        plot_list.append(ax3d.plot(x[mask], y[mask], mec='#77AC30', mfc=(0,0,0,0),\
                         marker=psm, ls=ls, **kwargs))
    
    mask = np.all((failsi, proxy), axis=0)
    if np.any(mask): #proxy_failures
        plot_list.append(ax3d.plot(x[mask], y[mask], mec='#D95319', mfc=(0,0,0,0),\
                         marker=pfm, ls=ls, **kwargs))
    
    return plot_list # success, failures, proxy_successes, proxy_failures



def tri_plot(ax, Tri=None, fmt='-', ns=100, **kwargs):
    if ax.sample_box.nvar == 3:
        lines = []
        
        if Tri is None:
            Tri = ax.sample_box.Tri
        
        # take coordinates in the space, where triangulation has been performed
        sampled_plan_tri = getattr(ax.sample_box, Tri.tri_space)
        
        #č mohli bychom zde machrovat se zkracenou smyčkou,
        #č my ale vůbec kontrolovat rovnost prostorů nebudeme, 
        #č nechť to dělá volající kód
        
        #if ax.space == Tri.tri_space:
        #   #return triplot(ax, **kwargs)
        #   
        #    #for simplex_id in simplex_ids:
        #    #    triangle = simplices[simplex_id]
        #    #    pos = sampled_plan_tri[triangle[[0,1,2,0]]]
             
        for simplex in Tri.tri.simplices:
            simplex_tri = sampled_plan_tri[simplex]
            x_tri_1 = np.linspace(simplex_tri[0,0], simplex_tri[1,0], ns, endpoint=False)
            y_tri_1 = np.linspace(simplex_tri[0,1], simplex_tri[1,1], ns, endpoint=False)
            z_tri_1 = np.linspace(simplex_tri[0,2], simplex_tri[1,2], ns, endpoint=False)
            
            x_tri_2 = np.linspace(simplex_tri[1,0], simplex_tri[2,0], ns, endpoint=False)
            y_tri_2 = np.linspace(simplex_tri[1,1], simplex_tri[2,1], ns, endpoint=False)
            z_tri_2 = np.linspace(simplex_tri[1,2], simplex_tri[2,2], ns, endpoint=False)
            
            x_tri_3 = np.linspace(simplex_tri[2,0], simplex_tri[3,0], ns, endpoint=True)
            y_tri_3 = np.linspace(simplex_tri[2,1], simplex_tri[3,1], ns, endpoint=True)
            z_tri_3 = np.linspace(simplex_tri[2,2], simplex_tri[3,2], ns, endpoint=True)
                
            
            tri_bound_tri = np.concatenate(((x_tri_1, y_tri_1, z_tri_1),\
                                            (x_tri_2, y_tri_2, z_tri_2),\
                                            (x_tri_3, y_tri_3, z_tri_3)), axis=1).T
            # vytvořme sample
            tri_bound = ax.sample_box.f_model.new_sample(tri_bound_tri, space=Tri.tri_space)
            
            xyz = getattr(tri_bound, ax.space)
            x, y, z = xyz.T
            lines.append(ax.plot(x, y, z, fmt, **kwargs))
        
        
            #
            #čs a eště raz
            #
        for simplex in Tri.tri.simplices:
            simplex_tri = sampled_plan_tri[simplex]
            x_tri_1 = np.linspace(simplex_tri[1,0], simplex_tri[3,0], ns, endpoint=False)
            y_tri_1 = np.linspace(simplex_tri[1,1], simplex_tri[3,1], ns, endpoint=False)
            z_tri_1 = np.linspace(simplex_tri[1,2], simplex_tri[3,2], ns, endpoint=False)
            
            x_tri_2 = np.linspace(simplex_tri[3,0], simplex_tri[0,0], ns, endpoint=False)
            y_tri_2 = np.linspace(simplex_tri[3,1], simplex_tri[0,1], ns, endpoint=False)
            z_tri_2 = np.linspace(simplex_tri[3,2], simplex_tri[0,2], ns, endpoint=False)
            
            x_tri_3 = np.linspace(simplex_tri[0,0], simplex_tri[2,0], ns, endpoint=True)
            y_tri_3 = np.linspace(simplex_tri[0,1], simplex_tri[2,1], ns, endpoint=True)
            z_tri_3 = np.linspace(simplex_tri[0,2], simplex_tri[2,2], ns, endpoint=True)
                
            tri_bound_tri = np.concatenate(((x_tri_1, y_tri_1, z_tri_1),\
                                            (x_tri_2, y_tri_2, z_tri_2),\
                                            (x_tri_3, y_tri_3, z_tri_3)), axis=1).T
        
            # vytvořme sample
            tri_bound = ax.sample_box.f_model.new_sample(tri_bound_tri, space=Tri.tri_space)
            
            xyz = getattr(tri_bound, ax.space)
            x, y, z = xyz.T
            lines.append(ax.plot(x, y, z, fmt, **kwargs))
        
        return lines




def limit_state_wireframe(ax3d, **kwargs):
    xmin, xmax = ax3d.get_xlim()
    ymin, ymax = ax3d.get_ylim()
    limits = (xmin, xmax, ymin, ymax)
    
    if 'rcount' in kwargs:
        ngrid = kwargs['rcount']
    else:
        ngrid = 50
    
    X, Y, Z = get_g_model_wireframe_data(ax3d.sample_box,\
                             limits, space=ax3d.space, ngrid=ngrid)
        
    return ax3d.plot_wireframe(X, Y, Z, **kwargs)


def limit_state_surface(ax3d, **kwargs):
    xmin, xmax = ax3d.get_xlim()
    ymin, ymax = ax3d.get_ylim()
    limits = (xmin, xmax, ymin, ymax)
    
    if 'rcount' in kwargs:
        ngrid = kwargs['rcount']
    else:
        ngrid = 50
    
    X, Y, Z = get_g_model_wireframe_data(ax3d.sample_box,\
                             limits, space=ax3d.space, ngrid=ngrid)
    
    # make a color map of fixed colors
    if 'cmap' not in kwargs:
        kwargs['cmap'] = colors.ListedColormap(['red', 'green'])
    if 'norm' not in kwargs:
        kwargs['norm'] = colors.BoundaryNorm([0], kwargs['cmap'].N, extend='both')
        
    return ax3d.plot_surface(X, Y, Z, **kwargs)


def density_surface(ax3d, **kwargs):
    xmin, xmax = ax3d.get_xlim()
    ymin, ymax = ax3d.get_ylim()
    
    if 'rcount' in kwargs:
        ngrid = kwargs['rcount']
    else:
        ngrid = 50
    
    x = np.linspace(xmin, xmax, ngrid)
    y = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(x, y)
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    z = ax3d.sample_box.sample_pdf(XY, ax3d.space)
    Z = z.reshape(ngrid, ngrid)
    
    return ax3d.plot_surface(X, Y, Z, **kwargs)
    

def density_colored_surface(ax3d, hull, colors=('#7735C2', '#808080'), **kwargs):
    xmin, xmax = ax3d.get_xlim()
    ymin, ymax = ax3d.get_ylim()
    
    inside_color, outside_color = colors
    
    if 'rcount' in kwargs:
        ngrid = kwargs['rcount']
    else:
        ngrid = 50
    
    x = np.linspace(xmin, xmax, ngrid)
    y = np.linspace(ymin, ymax, ngrid)
    X, Y = np.meshgrid(x, y)
    XY = np.vstack((X.reshape(-1), Y.reshape(-1))).T
    z = ax3d.sample_box.sample_pdf(XY, ax3d.space)
    Z = z.reshape(ngrid, ngrid)
    
    # for facecolors
    _x = np.linspace(xmin, xmax, ngrid-1, endpoint=False) + (xmax-xmin)/(ngrid-1)/2
    _y = np.linspace(ymin, ymax, ngrid-1, endpoint=False) + (ymax-ymin)/(ngrid-1)/2
    _X, _Y = np.meshgrid(_x, _y)
    _XY = np.vstack((_X.reshape(-1), _Y.reshape(-1))).T
    facecolors = np.full(len(_XY), inside_color)
    mask = hull.is_outside(hull.sample.f_model.new_sample(_XY, space=ax3d.space))
    facecolors[mask] = outside_color
    facecolors = facecolors.reshape(ngrid-1, ngrid-1)
    
    return ax3d.plot_surface(X, Y, Z, facecolors = facecolors, **kwargs)

    
def tri_surface(ax3d, **kwargs):
    xy = getattr(ax3d.sample_box, ax3d.space)
    x, y = xy.T
    z = ax3d.sample_box.g_values
    if 'zs' in kwargs:
        z = kwargs.pop('zs')
    else:
        z = ax3d.sample_box.pdf(ax3d.space)
    return ax3d.plot_trisurf(x, y, z, **kwargs)
    



# from Matplotlib sources:
# "TODO: Support custom face colours"
    
#def tri_colored_surface(ax3d, **kwargs):
#    xy = getattr(ax3d.sample_box, ax3d.space)
#    x, y = xy.T
#    z = ax3d.sample_box.g_values
#    tri = mtri.Triangulation(x, y)
#    colors = get_simplices_colors(ax3d.sample_box, tri.get_masked_triangles())
#    return ax3d.plot_trisurf(tri, z, color=colors, **kwargs)


def tri_colored_surfaces(ax3d, **kwargs):
    xy = getattr(ax3d.sample_box, ax3d.space)
    x, y = xy.T
    if 'zs' in kwargs:
        z = kwargs.pop('zs')
    else:
        z = ax3d.sample_box.g_values
    if 'tri_colors' in kwargs:
        tri_colors = kwargs.pop('tri_colors')
    else:
        tri_colors = ("xkcd:light seafoam green", "xkcd:pale rose", "xkcd:dark cream")
    tri = mtri.Triangulation(x, y)
    triangles = tri.get_masked_triangles()
    events = get_events(ax3d.sample_box, triangles)
    # 0=success, 1=failure, 2=mix
    surf = list()
    if np.any(events==0):
        surf.append(ax3d.plot_trisurf(x, y, triangles[events==0], z, color=tri_colors[0], **kwargs))
    if np.any(events==1):
        surf.append(ax3d.plot_trisurf(x, y, triangles[events==1], z, color=tri_colors[1], **kwargs))
    if np.any(events==2):
        surf.append(ax3d.plot_trisurf(x, y, triangles[events==2], z, color=tri_colors[2], **kwargs))
    return surf



def wall(ax3d, lines2d, zmin=0, zmax=1, **kwargs):
    surf = list()
    for line2d in lines2d:
        x = line2d.get_xdata()
        y = line2d.get_ydata()
        _x = np.concatenate((x, x))
        _y = np.concatenate((y, y))
        _z = np.concatenate([np.full(len(x), zmin), np.full(len(x), zmax)])
        tri = mtri.Triangulation(_x, _z)
        triangles = tri.get_masked_triangles()
        surf.append(ax3d.plot_trisurf(_x, _y, _z, triangles=triangles, **kwargs))
    
    



#!/usr/bin/env python
# coding: utf-8

"""
SampleBox = sample_R(f_model) + g_values
"""


import numpy as np


class SampleBox:
    """
    SampleBox = sample_R(f_model) + g_values
    
    .sampled_plan object
    .g_values
    .failsi
    
    Souřadnice primárně z prostoru modelu, ty co jsme rovnou
    posilali do g_modelu!
    """
    
        # nechtěl bys nazvy proměnných?
    def __new__(cls, sample_object, g_values=(), gm_signature=''):
        """
        Jedname tvrdě - není-li vstup konzistentní, 
        tak sbox vůbec nevytvaříme
        """
        g_values = np.atleast_1d(g_values)
        if len(sample_object) == len(g_values):
            sb = super(SampleBox, cls).__new__(cls)
            # nepotrebujeme žádné rozdělení, nic
            sb.sampled_plan = sample_object
            sb.g_values = g_values
            sb.gm_signature = gm_signature
            return sb
        else:
            raise ValueError("Sample and g_value hasn't the same length. Zkrátka, do sebe nepatří")
    
        
    def __str__(sb):
        return  '%s: %s at %s' %(sb.gm_signature, sb.g_values, sb.sampled_plan)
        
    def __repr__(sb):
        return  'SampleBox(%s, %s, %s)' %(repr(sb.sampled_plan), repr(sb.g_values), repr(sb.gm_signature))
        
    def __len__(sb):
        return len(sb.g_values)
        
        
    def __call__(sb):
        # я ваще хз
        # offer next sample?
        # do calculation?
        # add to this sample?
        # return new instance?
        # мар, сакра, кароно?
        
        # finally, we will offer sample to sample
        # like BlackBox does
        return sb.sampled_plan(1)
    
        
    def __getitem__(sb, slice):
        return SampleBox(sb.sampled_plan[slice], sb.g_values[slice], sb.gm_signature)
        
        
    def __getattr__(sb, attr):
        if attr == 'samplebox':
            return sb
        elif attr == 'failsi':
            # ~(g_values>0) to handle nan
            return ~(sb.g_values>0)
        elif attr == 'success_points':
            return np.argwhere(sb.g_values>0).reshape(-1)
        elif attr == 'failure_points':
            return np.argwhere(~(sb.g_values>0)).reshape(-1)
        elif attr == 'failure_samples':
            return sb[~(sb.g_values>0)]
        elif attr == 'success_samples':
            return sb[sb.g_values>0]
            
        # to je jistě k samplovi
        else:
            return getattr(sb.sampled_plan, attr)
        
        
    def add_sample(sb, input_sb):
        input_sb.consistency_check()
        
        # ты чьих будешь?
        # where are you from?
        # are you one of us?
        if sb.gm_signature == input_sb.gm_signature:
            # dá se tuhle kontrolu jednoduše napálit, ale to neřeším
            sb.sampled_plan.add_sample(input_sb.sampled_plan)
            sb.g_values = np.append(sb.g_values, input_sb.g_values)
            
            return sb.consistency_check()
            
            # je to pro případ prázdného sample_boxu
        elif sb.gm_signature == '':
            # dá se tuhle kontrolu jednoduše napálit, ale to neřeším
            sb.sampled_plan.add_sample(input_sb.sampled_plan)
            sb.g_values = np.append(sb.g_values, input_sb.g_values)
            sb.gm_signature = input_sb.gm_signature
            return sb.consistency_check()
        else:
            #raise ValueError("Merge sa nám nějak nepovedol")
            print(sb.gm_signature, input_sb.gm_signature)
            print(type(sb.gm_signature), type(input_sb.gm_signature))
            raise ValueError("gm_signatures are unequal. You are probably trying to merge data from different sources")
                
    def new_sample(sb, input_sb):
        """
        We want to create new SampleBox object with our distribution (f_model)
        but with data of input_sb (just like f_model.new_sample() does)
        """
        return SampleBox(sb.sampled_plan.new_sample(input_sb), input_sb.g_values, input_sb.gm_signature)
        
    def consistency_check(sb):
        if len(sb.sampled_plan)==len(sb.g_values):
            return True
        else:
            # уг тодӥськы чик мар кароно
            # ConsistencyError
            raise ValueError('SampleBox is in an inconsistent state and nobody knows what to do with it')
            



#!/usr/bin/env python
# coding: utf-8

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtWidgets
from pyqtgraph.Qt import QtCore

import numpy as np

from ..voronoi import ContactVoronoi

CV = None

##č (jednotka dočasnosti - jeden furt)

class ContactVoronoiBaseWidget(pg.LayoutWidget):
	#mixed_added = QtCore.pyqtSignal(int)
	#mixed_updated = QtCore.pyqtSignal(int)
	#mixed_deleted = QtCore.pyqtSignal(int)
	
	def gen_cv(self): 
		
		model_space = self.param.child('model_space').value()
		ns = self.param.child('ns').value()
		p_base = self.param.child('p_base').value()
		workers = self.param.child('workers').value()
		global CV = ContactVoronoi(self.w.get_sample_box(), self.get_hull(), 
						model_space=model_space, ns=ns, p_base=p_base,
						auto_update=True, workers=workers)
#						on_add_mixed=self.add_mixed_callback,
#						on_update_mixed=self.update_mixed_callback, 
#						on_delete_mixed=self.delete_mixed_callback)
			
	def get_cv(self): 
		if CV is None:
			self.gen_cv()	
		return CV
			
			


class VoronoiEstimationWidget(QtWidgets.QSplitter):
    """
    addLabel(text=' ', row=None, col=None, rowspan=1, colspan=1, **kargs)
    """
    # I'd like to get access to the samplebox stuff via the container's reference, 
    # INHERETED by gl_plot
    def __init__(self, samplebox_item,  parent=None, *args, **kwargs):
        super().__init__(parent)
        # sb like samplebox, of course
        self.sb_item = samplebox_item
        
        self.sb_item.box_runned.connect(self.on_box_run)
        self.sb_item.slice_changed.connect(self.self_clear)
        self.sb_item.space_changed.connect(self.on_space_changed)
        self.sb_item.redraw_called.connect(self.redraw)
        #☺ na internetu všichni tak dělaj
        self.setup()
    
    # INHERETED by gl_plot
    def setup(self):
        self.setOrientation(QtCore.Qt.Vertical)
        self.layout = pg.LayoutWidget(self)
        # model_space='Rn', sampling_space=None, p_norm=1, budget=20000
        params = [{'name': 'method', 'type': 'list', 'values': ['Voronoi_tesselation','Voronoi_2_point_estimation'], 'value': 'Voronoi_2_point_estimation'}]
        params.append({'name': 'model space', 'type': 'list', 'values': self.sb_item.spaces, 'value': 'Rn'})
        params.append({'name': 'sampling space', 'type': 'list', 'values': ['None'] + self.sb_item.spaces, 'value': 'None'})
        params.append({'name': 'p-norm', 'type': 'float', 'limits': (1, float('inf')), 'value': 1, 'default': np.inf})
        params.append({'name': 'budget', 'type': 'float', 'limits': (1, float('inf')), 'value': 20000, 'default': 20000})
        self.coloring_modes = ['simple_coloring', 'cell_probability_coloring','node_pf_coloring']
        params.append({'name': 'coloring', 'type': 'list', 'values': self.coloring_modes, 'value': self.coloring_modes[1]})
        params.append({'name': 'node (pixel) size', 'type': 'float', 'limits': (0, float('inf')), 'value': 3, 'default': 1})
        params.append({'name': 'Run with the box', 'type': 'bool', 'value': False }) # 'tip': "This is a checkbox"
        
        ### Create tree of Parameter objects
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        
        ### Create ParameterTree widget
        self.ptree = pg.parametertree.ParameterTree()
        self.ptree.setParameters(self.param, showTop=False)
        
        self.layout.addWidget(self.ptree, row=0, col=0, colspan=3)
        
        
        self.btn = QtWidgets.QPushButton('estimate')
        self.layout.addWidget(self.btn, row=1, col=0)
        self.btn.clicked.connect(self.run_stm)
        
        self.btn2 = QtWidgets.QPushButton('redraw')
        self.layout.addWidget(self.btn2, row=1, col=1)
        self.btn2.clicked.connect(self.recolor)
        
        self.btn3 = QtWidgets.QPushButton('hide')
        self.layout.addWidget(self.btn3, row=1, col=2)
        self.btn3.clicked.connect(self.hide)
        
        self.addWidget(self.layout)
        
        self.table = pg.TableWidget(sortable=False)
        self.addWidget(self.table)
        
        # pro začatek postačí
        self.cells = []
        # probability of the biggest cell
        # used for coloring
        self.p_cell_max = {'success':0, 'failure':0}
     
     
    # INHERETED by gl_plot
    def on_box_run(self, *args, **kwargs):
        # je třeba zkontrolovat autorun a restartovat výpočet
        if self.param.getValues()['Run with the box'][0]:
            self.run_stm()
        #else:
            #self.self_clear()
    
    # INHERETED by gl_plot
    def hide(self):
        for nodes, plot_item, cell_stats in self.cells:
            plot_item.hide()
            # keep the GUI responsive :)
            #self.sb_item.app.processEvents()
    
    # INHERETED by gl_plot
    def redraw(self, *args, **kwargs):
        self.cells.clear()
        self.p_cell_max['success'] = 0
        self.p_cell_max['failure'] = 0
    
    ## I'll rename after main widget refactoring
    # refactoring already done, why I should rename?
    # INHERETED by gl_plot
    def recolor(self):
        # indikace
        #self.setDisabled(True)
        with pg.BusyCursor():
            # Krucinal, kdo ten OrderedDict vymyslel?
            params = self.param.getValues()
            coloring = params['coloring'][0]
            
            # přebarvíme nějak tečičky 
            # callback vybírám ze svého kódu, ten musí bejt v pořádku
            # ne že by to bylo dokonalé bezpečný, 
            # ale na lokálním počítači asi to není až tak zavadný
            coloring_function = getattr(self, coloring)
            
            # hura! Jedeme!
            coloring_function()
            
            
        # indikace
        #self.setEnabled(True)
    
    
    def on_space_changed(self, *args, **kwargs):
        # teď tečičky
        for nodes, plot_item, cell_stats in self.cells:
            pos = getattr(nodes, self.sb_item.space)[:,:2]
            plot_item.setData(pos)
        
        
    # INHERETED by gl_plot
    def self_clear(self):
        # odebereme prvky-propísky z hlavního plotu
        for nodes, plot_item, cell_stats in self.cells:
            self.sb_item.central_widget.removeItem(plot_item)
        
        self.redraw()
        


    # INHERETED by gl_plot
    def run_stm(self):
        # indikace
        #self.setDisabled(True)
        with pg.BusyCursor():
            nsim = self.sb_item.slider.value()
            sample_box = self.sb_item.sample_box[:nsim]
            # Krucinal, kdo ten OrderedDict vymyslel?
            params = self.param.getValues()
            method = params['method'][0]
            model_space = params['model space'][0]
            sampling_space = params['sampling space'][0]
            if sampling_space == 'None':
                sampling_space = None
            p_norm = params['p-norm'][0]
            budget = params['budget'][0]
            coloring = params['coloring'][0]
            
            # je třeba skrýt prvky z minula
            self.self_clear()
            
            
            # přebarvíme nějak tečičky 
            # callback vybírám ze svého kódu, ten musí bejt v pořádku
            # ne že by to bylo dokonalé bezpečný, 
            # ale na lokálním počítači asi to není až tak zavadný
            coloring_function = getattr(self, coloring)
            
            
            try:
                stm_function = getattr(stm, method)
                # model_space='Rn', sampling_space=None, p_norm=1, budget=20000
                data = stm_function(sample_box, model_space=model_space, sampling_space=sampling_space,\
                                                     p_norm=p_norm, budget=budget, callback=coloring_function)
                
                if hasattr(self.sb_item.sample_box, 'estimations'):
                    self.sb_item.sample_box.estimations.append(data)
                    self.sb_item.estimation_added.emit()
                self.table.setData(data)
                    
            except BaseException as e:
                msg = "error during estimation "
                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
                print(error_msg)
                
            
            
        # indikace
        #self.setEnabled(True)
    
        
        
            
        
    def node_pf_coloring(self, estimation=None, nodes=None, cell_stats=None, out_nodes=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        plot_widget = self.sb_item.central_widget
        if nodes is None:
            for cell in self.cells:
                nodes, plot_item, cell_stats = cell
                # odebereme prvky z hlavního plotu
                # zde je třeba prvky vygenerovat znovu 
                # protože nikdo neví co tam bylo před tím
                # takhle, nechce se mi drbat s tím, co tam bylo před tím
                # komplikace ze strany pyqtgraph
                plot_widget.removeItem(plot_item)
                
                # bacha, potřebuji prvek uložiť in-place
                cell[1] = self.node_pf_scatter_plot(nodes, cell_stats)
        
        # máme nodes, tj. jedeme poprvé        
        else:
            plot_item = self.node_pf_scatter_plot(nodes, cell_stats)
            
            # uložíme data
            self.cells.append([nodes, plot_item, cell_stats])
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
            
            
            
            
    def node_pf_scatter_plot(self, nodes, cell_stats):
        pos = getattr(nodes, self.sb_item.space)[:,:2]
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        plot_widget = self.sb_item.central_widget
        # zas, нет ножек - нет мультиков
        # node_pf_estimations nemusejí bejt
        try:
            # zkusmě pro jednoduchost 
            # čírou RGB hračku
            npf = nodes.node_pf_estimations
            colors = tuple((npf[i]*255, (1-npf[i])*255, 0) for i in range(len(pos))) 
            # sice dokumentace popisuje víc možností zadávání, 
            # ale toto zadávání různejch barviček je pro mě jediné fungujicí. Drbal jsem s tím do znechucení
            # je v podstatě opsané z příkladu
            # rovnou přes PlotDataItem mi nefunguje
            # žádné jiné možností zadávání já jsem v zdrojacích 
            # pyqtgraph (konkretně v PlotDataItem a v ScatterPlotItem) neuviděl
            # tuším, že je to neunosně drahý
            list_of_dicts = list({'pos': pos[i], 'size':symbol_size, 'pen': colors[i], 'brush':colors[i], 'symbol':'o'} for i in range(len(pos)))
            plot_item = pg.ScatterPlotItem(list_of_dicts)
            plot_widget.addItem(plot_item)
            return plot_item
            
        except BaseException as e:
            msg = "node_pf_coloring has problems "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)
            #self.error.emit(error_msg)
            # simple coloring
            event = cell_stats['event']
            color = self.get_color(event)
            #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
            return plot_widget.plot(pos, pen=None, symbol='o', symbolPen=color, symbolBrush=color, symbolSize=symbol_size, name='IS localized nodes')
        

        
        
    
    def simple_coloring(self, nodes=None, cell_stats=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        plot_widget = self.sb_item.central_widget
        if nodes is None:
            for cell in self.cells:
                nodes, plot_item, cell_stats = cell
                # odebereme prvky z hlavního plotu
                # zde je třeba prvky vygenerovat znovu 
                # protože nikdo neví co tam bylo před tím
                # takhle, nechce se mi drbat s tím, co tam bylo před tím
                # komplikace ze strany pyqtgraph
                plot_widget.removeItem(plot_item)
                
                # draw 
                pos = getattr(nodes, self.sb_item.space)[:,:2]
                #x, y = (*getattr(nodes, self.sb_item.space).T,)
                
                event = cell_stats['event']
                color = self.get_color(event)
                #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
                # bacha, potřebuji prvek uložiť in-place
                cell[1] = plot_widget.plot(pos, pen=None, symbol='o',\
                        symbolPen=color, symbolBrush=color, symbolSize=symbol_size, name='IS localized nodes')
        
        # máme nodes, tj. jedeme poprvé        
        else:
            # draw tečičky
            #
            pos = getattr(nodes, self.sb_item.space)[:,:2]
            
            event = cell_stats['event']
            color = self.get_color(event)
            #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
            plot_item = plot_widget.plot(pos, pen=None, symbol='o',\
                     symbolPen=color, symbolBrush=color, symbolSize=symbol_size, name='IS localized nodes')
            
            # uložíme data
            self.cells.append([nodes, plot_item, cell_stats])
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
    
        
            
    def cell_probability_coloring(self, nodes=None, cell_stats=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        plot_widget = self.sb_item.central_widget
        if nodes is None:
            # odebereme prvky z hlavního plotu
            # zde je třeba prvky vygenerovat znovu 
            # protože nikdo neví co tam bylo před tím
            # takhle, nechce se mi drbat s tím, co tam bylo před tím
            # komplikace ze strany pyqtgraph
            for nodes, plot_item, cell_stats in self.cells:
                plot_widget.removeItem(plot_item)
                
                event = cell_stats['event']
                cell_probability = cell_stats['cell_probability']
                if self.p_cell_max[event] < cell_probability:
                    self.p_cell_max[event] = cell_probability
                
            # přebarvíme tečičky podle obsahu pravděpodobnosti
            for cell in self.cells:
                nodes, plot_item, cell_stats = cell
                # draw 
                pos = getattr(nodes, self.sb_item.space)[:,:2]
                #x, y = (*getattr(nodes, self.sb_item.space).T,)
                
                event = cell_stats['event']
                cell_probability = cell_stats['cell_probability']
                # bez modrého - maximální intenzita
                blue_intensity = 1 - cell_probability / self.p_cell_max[event]
                color = self.get_color(event, blue_intensity)
                #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
                # bacha, potřebuji prvek vložit zpätky
                cell[1] = plot_widget.plot(pos, pen=None, symbol='o',\
                            symbolPen=color, symbolBrush=color, symbolSize=symbol_size, name='IS localized nodes')
        
        # máme nodes, tj. jedeme poprvé        
        else:
            event = cell_stats['event']
            cell_probability = cell_stats['cell_probability']
            # zkontrolujeme probability
            if self.p_cell_max[event] < cell_probability:
                self.p_cell_max[event] = cell_probability
                # a hned všecko dotyčné přebarvíme podle obsahu pravděpodobnosti
                for cell in self.cells:
                    if event == cell[2]['event']:
                        # bez modrého - maximální intenzita
                        # zde cell_probability se rovná self.p_cell_max[event]
                        # ale bacha! Nesplet se jinde!
                        blue_intensity = 1 - cell[2]['cell_probability'] / cell_probability
                        color = self.get_color(event, blue_intensity)
                        # bacha, potřebuji prvek vložit zpätky
                        cell[1].setSymbolBrush(color)
                        cell[1].setSymbolPen(color)
            
            # bez modrého - maximální intenzita
            blue_intensity = 1 - cell_probability / self.p_cell_max[event]
            color = self.get_color(event, blue_intensity)
            
            
            # draw tečičky
            #
            pos = getattr(nodes, self.sb_item.space)[:,:2]
            #x, y = (*getattr(nodes, self.sb_item.space).T,)
            
            #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
            plot_item = plot_widget.plot(pos, pen=None, symbol='o',\
                        symbolPen=color, symbolBrush=color, symbolSize=symbol_size, name='IS localized nodes')
            
            # uložíme data
            self.cells.append([nodes, plot_item, cell_stats])
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
        
                
                
    def get_color(self, event, blue_intensity=None):
        """
        get color for 'simple_coloring' or 'cell_probability_coloring'
        """
        # generally
        if event == 'success':
            color = [167, 255, 181]# xkcd:light seafoam green #a7ffb5
        elif event == 'failure':
            color = [253, 193, 197] # xkcd: pale rose (#fdc1c5)
        # já vím, že Voronoi nemá 'mix', эн но юа
        else:# 'mix'
            color = [255, 243, 154] # let's try xkcd: dark cream (#fff39a)
            
        if blue_intensity is not None:
            # let's play with blue a little bit
            # chcu mít korektní výstup
            # aspoň v něčem chcu být jistý
            if blue_intensity > 1:
                color[2] = 255
            elif blue_intensity < 0:
                color[2] = 0
            else:
                # pyqtgraph žere barvy i s čarkou
                # ale my je mu davat nebudeme
                color[2] = int(blue_intensity*255) 
            
        return tuple(color)            


class ContactVoronoiBaseWidget(pg.LayoutWidget):
    # I'd like to get access to the samplebox stuff via the container's reference, 
    def __init__(self, samplebox_item,  parent=None, *args, **kwargs):
        super().__init__(parent)
        # sb like samplebox, of course
        self.sb_item = samplebox_item
        
        self.giracle = Giracles(w=samplebox_item, autoredraw=False, nrod=200)
        #č Ghull se zkomplikoval. Musím integraci řešit zvlášť
        # Serie for integration nodes
        self.serint = Series(w=samplebox_item, autoredraw=False)
        #ё hyperplanes? Кого ты обманываешь, Alexi?
        #č tak. už je to equation_planes :)
        self.equation_planes = InfiniteLines(w=samplebox_item, autoredraw=False)
        
        # signals handling:
        # box_runned - handled by .on_box_run() method
        # space_changed - handled automatically by smart items
        # slice_changed - setted up clear() on smart items
        # redraw_called - handled automatically by smart items
        # estimation_added - class does not handle the signal
        #                   nor emits the signal itself 
        #                   (as does not save estimations anywhere for now)
        
        # todo: design estimation record 
        
        self.sb_item.box_runned.connect(self.on_box_run)
        #č Alexi, ten signal může té funkce posilát bůhví co navíc.
        #č (ano. clear() nebere žádné argumenty, takže v cajku)
        self.sb_item.slice_changed.connect(self.giracle.clear)
        self.sb_item.slice_changed.connect(self.equation_planes.clear)
        #č ty naše chytré prvky giracle a equation_planes 
        #č hlídají space_changed a redraw_called sami.
        
        self.schemes = dict()
        self.ndim = self.sb_item.sample_box.nvar
        #☺ na internetu všichni tak dělaj
        self.setup()
        
    def setup(self):
        #č já teďkom dědím ne QSplitter, ale LayoutWidget
        #оӵ Кужым кариськы, Олёш!
        
        #č zkusíme nový (pro WellMet) UI, uživatelské rozhraní
        #оӵ кнопка вылын
        #self.layout = self.addLayout(row=0, col=0)
        #self.tool_layout = QtGui.QHBoxLayout()
        #self.layout.addLayout(self.tool_layout, 0, 0)
        #self.addWidget(self.layout, row=0, col=0)
        
        #č toolbar je obecně super věc, ale mě zlobí umístění v layoutu
        self.toolbar = QtWidgets.QToolBar(self)
        #č jmenovitě, roztažení mě nefunguje vůbec
        #size_policy = self.toolbar.sizePolicy()
        #size_policy.setHorizontalPolicy(QtGui.QSizePolicy.Expanding)
        #self.toolbar.setSizePolicy(size_policy)
        
        #č draw_convex_hull ja navržena tak, aby brala jíž hotový hull
#        if self.ndim == 2:
#            action = self.toolbar.addAction("draw convex hull", self.draw_convex_hull)
#            btn = self.toolbar.widgetForAction(action)
#            btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("shell out!", self.get_shell_estimation)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Creates Ghull object")
        #btn.setSizePolicy(size_policy)
        
        action = self.toolbar.addAction("integrate", self.integrate)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Only uses Ghull object created before")
        #btn.setSizePolicy(size_policy)
        
        action = self.toolbar.addAction("shot!", self.shot)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("fire!", self.fire)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        #btn.setSizePolicy(size_policy)
        
        action = self.toolbar.addAction("boom!", self.boom)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("hide", self.hide)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        #btn.setSizePolicy(size_policy)
        
        action = self.toolbar.addAction("show", self.show)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("clear", self.clear)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        #self.tool_layout.addWidget(self.toolbar)
        self.addWidget(self.toolbar, row=0, col=0)
        
        
        #оӵ остальной (люкет) уллапала
        
        ### Create ParameterTree widget
        self.ptree = pg.parametertree.ParameterTree()
        self._set_param()
        self.ptree.setParameters(self.param, showTop=False)
        
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.addWidget(self.ptree)
        
        self.table = pg.TableWidget(sortable=False)
        self.splitter.addWidget(self.table)
        
        #self.addWidget(self.splitter, row=1, col=0, colspan=5)
        self.addWidget(self.splitter, row=1, col=0)
    
    def hide(self):
        self.serint.hide()
        self.giracle.hide()
        self.equation_planes.hide()
        
    def show(self):
        self.serint.show()
        self.giracle.show()
        self.equation_planes.show()
        
    def clear(self):
        self.serint.clear()
        self.giracle.clear()
        self.equation_planes.clear()
    
    #оӵ DirectHull понна гинэ
    #č pouze pro DirectHull
    def get_scheme(self):
        scheme = self.param.getValues()['scheme'][0]
        if scheme == 'random':
            ndir = self.param.getValues()['ndir'][0]
            return sball.get_random_directions(ndir, self.ndim)
        elif scheme in self.schemes:
            return self.schemes[scheme].points
        else:
            Scheme = getattr(quadpy.un, scheme)
            self.schemes[scheme] = Scheme(self.ndim)
            return self.schemes[scheme].points

        
    def _set_param(self):
        params = list()
        params.append({'name': 'method', 'type': 'list', 'value': 'DirectHull', \
                     'values': ['SBall', 'BrickHull', 'DirectHull', 'CompleteHull', 'QHull', 'Grick']})
        params.append({'name': 'space', 'type': 'list', 'tip': "Not used for SBall and Grick", \
                    'values': self.sb_item.spaces, 'value': 'G'})
                    
        schemes_list = schemes.un_spheres + ['random']
        params.append({'name': 'scheme', 'type': 'list', \
                     'values': schemes_list, 'value': schemes_list[0], \
                     'tip': "Used only for Grick, DirectHull and CompleteHull. Generation can take for a while"})
        params.append({'name': 'ndir', 'type': 'int', \
                    'limits': (1, float('inf')), 'value': 1000, 'default': 1000, \
                    'title': "number of random directions", \
                    'tip': "Used only for Grick or for random scheme in DirectHull (or CompleteHull)"})   
            
        params.append({'name': 'integrator', 'title': "integrator", 'type': 'list', \
                    'values': ['MC', 'IS', '1DS'], 'value': '1DS' })
        params.append({'name': 'nonG_reduction', 'type': 'float', \
                    'title': "non Gaussian reduction", \
                    'limits': (0, 1), 'value': 0.9, 'default': 0.9,\
                    'tip': "Applied when Ghull integrates non Gaussian convex hulls"})   
        
        params.append({'name': 'use_MC', 'title': "use MC", 'type': 'bool', 'value': False, \
                    'tip': "Used for shot(), fire() and boom() functions"})   
        
        params.append({'name': 'budget', 'type': 'int',  \
                    'limits': (1, float('inf')), 'value': 1000, 'default': 1000,\
                    'tip': "Number of simulations for optimal importance sampling"})   
                    
        params.append({'name': 'node (pixel) size', 'type': 'float',\
                 'limits': (0, float('inf')), 'value': 3.5, 'default': self.sb_item.px_size})
        #ё больше калорий богу калорий!
        params.append({'name': 'r', 'type': 'color', 'value': (189, 204, 0, 255) })
        params.append({'name': 'inside', 'type': 'color', 'value': (133, 172, 102, 255) })
        params.append({'name': 'convex_hull', 'type': 'color', 'value': (85, 170, 255, 255) }) # (186, 109, 0, 255)
        params.append({'name': 'fire', 'type': 'color', 'value': (245, 117, 0, 255) })
        params.append({'name': 'orth', 'type': 'color', 'value': (255, 0, 0, 255) })
        params.append({'name': 'outside', 'type': 'color', 'value': 0.6})
        params.append({'name': 'R', 'type': 'color', 'value': (85, 85, 255, 255) })
        params.append({'name': 'Update as the box runned', 'type': 'bool', 'value': False }) # 'tip': "This is a checkbox"
        params.append({'name': 'index', 'title': "replace previous", 'type': 'bool', 'value': True })
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
    
    def get_ghull(self): 
        integrator = self.param.getValues()['integrator'][0]
        if integrator == 'MC':
            Integrator = Shell_MC
        elif integrator == 'IS':
            Integrator = Shell_IS
        elif integrator == '1DS':
            Integrator = Shell_1DS
        nonG_reduction = self.param.getValues()['nonG_reduction'][0]
        hull = self.get_hull()
        self.ghull = Ghull(hull, Integrator=Integrator, non_Gaussian_reduction=nonG_reduction)
        return self.ghull
                        
        
    #č jistě potřebuji něco, co zpracuje parameter_tree 
    #č a vratí platný hull pro Ghull
    def get_hull(self):
        #č bez semplu se neobejde
        nsim = self.sb_item.slider.value()
        sample = self.sb_item.sample_box.f_model[:nsim]
        
        # ['SBall', 'BrickHull', 'DirectHull', 'QHull', 'Grick']
        hull_model = self.param.getValues()['method'][0]
        if hull_model == 'SBall':
            return khull.GBall(sample)
        elif hull_model == 'BrickHull':
            space = self.param.getValues()['space'][0]
            return khull.BrickHull(sample, space)
        elif hull_model == 'DirectHull':
            space = self.param.getValues()['space'][0]
            direct_plan = self.get_scheme()
            return khull.DirectHull(sample, direct_plan, space)
        elif hull_model == 'CompleteHull':
            space = self.param.getValues()['space'][0]
            direct_plan = self.get_scheme()
            return khull.CompleteHull(sample, direct_plan, space)
        elif hull_model == 'QHull':
            space = self.param.getValues()['space'][0]
            #č tento widget pokažde generuje obálku znovu
            return khull.QHull(sample, space, incremental=False)
        elif hull_model == 'Grick':
            direct_plan = self.get_scheme()
            ndir = self.param.getValues()['ndir'][0]
            return khull.Grick(sample, direct_plan, nrandom=ndir, auto_update=True)
        else:
            raise ValueError("HullEstimationWidget: co to je za obálku?")

    #č ten hlavní modul se dočkal na překopávání
    def on_box_run(self, *args, **kwargs):
        self.clear()
        #č je třeba zkontrolovat autorun a restartovat výpočet
        if self.param.getValues()['Update as the box runned'][0]:
            self.get_shell_estimation()
    
    def index(self, index):
        if self.param.getValues()['index'][0]: # replace previous
            return index
        else:
            return None
    
    def draw_planes(self, equations, space, **kwargs):
        if self.ndim == 2:
            #č musíme něco zavolat na self.equation_planes
            #č equation_planes má funkci add_line()
            #č add_line(self, space='G', index=None, **plot_kwargs)
            #č která pak plot_kwargs přeposilá funkci addLine()
            #č na central widgetu. 
            #č To vše skončí ve pyqtgrafové InfiniteLine třidě.
            #č ta moje třida InfiniteLines sama se stará o shodování prostorů
            #č indexy posilat nebudeme (s nimi je to trošku komplikovanější)
            
            #pos = list() #č navrhové body nakreslíme všechny dohromady
            for equation in equations:
                #č ve 2D bych očekával v rovnici pouze 3 hodnoty (já potřebuji směry)
                x, y, offset = equation
                design_point = [-x*offset, -y*offset]
                #self.sb_item.central_widget.plot(np.array([pos, pos]), symbol='o')
    #                    if y < 0: #č tak to aspoň kreslí
    #                        angle = np.rad2deg(np.arcsin(x))
    #                    else:
    #                        angle = np.rad2deg(np.arccos(y))
                    
                if (x*y) < 0: #č tak to aspoň kreslí
                    angle = np.rad2deg(np.arccos(np.abs(y)))
                else:
                    angle = np.rad2deg(np.arccos(-np.abs(y)))
                self.equation_planes.add_line(space=space, pos=design_point, angle=angle, **kwargs)
                
                
    
    def draw_convex_hull(self, hull):
        try:
            if self.param.getValues()['index'][0]: # replace previous
                self.equation_planes.clear()
            
            #č zatím uděláme jen pro 2D infinite lajny
            design_points = hull.get_design_points()
            
            size = self.param.getValues()['node (pixel) size'][0]
            color = self.param.getValues()['convex_hull'][0] #č tam bude barva
            
            self.giracle.add_serie(design_points, z=31, index=self.index('design points'),\
                            pen=None, symbol='o', symbolPen=pg.mkPen(None), \
                            symbolBrush=color, symbolSize=size, name='design points')
            self.draw_planes(hull.equations, space=hull.space, z=29,  pen=color)
                            
                            
            # orth
            color = self.param.getValues()['orth'][0] #č tam bude barva
            #self.giracle.add_serie(FORM_points, z=32, index=self.index('2FORM points'),\
            #                pen=None, symbol='o', symbolPen=pg.mkPen(None), \
            #                symbolBrush=color, symbolSize=size, name='2FORM points')
            self.draw_planes(hull.get_orth_equations(), space=hull.space, z=30,  pen=color)
            
            
            # 2FORM
            color = self.param.getValues()['fire'][0] #č tam bude barva
            #self.giracle.add_serie(FORM_points, z=32, index=self.index('2FORM points'),\
            #                pen=None, symbol='o', symbolPen=pg.mkPen(None), \
            #                symbolBrush=color, symbolSize=size, name='2FORM points')
            self.draw_planes(hull.get_2FORM_equations(), space=hull.space, z=30,  pen=color)
                            
                            
        except BaseException as e:
            msg = "draw_convex_hull error "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)
            self.sb_item.errors.append(e)
           
    def draw_ghull(self, r, R):
        # r-circle
        if r < 0:
            r = 0
        color = self.param.getValues()['r'][0] #č tam bude barva
        #č krucí, nevím co mám dělat s indexama. 
        #č co mám dělat s předchozí kresbou? Nechame
        self.giracle.add_circle(r=r, index=self.index('r'),z=32, pen=color, name='r')
        
        # R-circle. #č Při kreslení nahradíme předchozí.
        color = self.param.getValues()['R'][0] #č tam bude barva
        self.giracle.add_circle(R, z=32, index=self.index('R'), pen=color, name='R')
            
        
        
    
    def get_shell_estimation(self):
        ghull = self.get_ghull()
        self.draw_convex_hull(ghull.hull)

        try:
            shell_estimation, global_stats = ghull.get_shell_estimation()
            
#            if hasattr(self.sb_item.sample_box, 'estimations'):
#                self.sb_item.sample_box.estimations.append(data)
#                self.sb_item.estimation_added.emit()
            self.table.setData({**global_stats, 'shell_estimation':shell_estimation})
            
            self.draw_ghull(global_stats['r'], global_stats['R'])
            
        except BaseException as e:
            msg = "error during estimation "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)
            self.sb_item.errors.append(e)
        
        
    def shot(self):
        ghull = self.get_ghull()
        self.draw_candidates(ghull.hull.shot)
        
    def fire(self):
        ghull = self.get_ghull()
        self.draw_candidates(ghull.hull.fire)
        
    def boom(self):
        ghull = self.get_ghull()
        self.draw_candidates(ghull.boom)
    
    def draw_candidates(self, source_function):
        #č špatně rozumím, co tím bylo mysleno
        #č že jsem chtěl kreslit náhodné směry?
        ##č zatím máme issue s náhodným planem
        ##č kdyby něco - vyřeším přes ukladání 
        ##č náhodných plánů v parameter_tree
        
        self.draw_convex_hull(self.ghull.hull)
        
        #☺ Krucinal, kdo ten OrderedDict vymyslel?
        params = self.param.getValues()
        ns = params['budget'][0]
        use_MC = params['use_MC'][0]
        try:
            fire = source_function(ns, use_MC=use_MC)
            # draw tečičky
            color = self.param.getValues()['fire'][0] #č tam bude barva
            size = self.param.getValues()['node (pixel) size'][0]
            #brush = pg.mkBrush(color)
            self.giracle.add_serie(fire, z=30, index=self.index('fire'),\
                        pen=None, symbol='o', symbolPen=pg.mkPen(None), \
                        symbolBrush=color, symbolSize=size, name='fire')

        except BaseException as e:
            msg = "error "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)
            self.sb_item.errors.append(e)        
    
    def integration_callback(self, nodes):
        #č draw tečičky
        size = self.param.getValues()['node (pixel) size'][0]
        plot_params = {'pen':None, 'symbol':'o', 'symbolSize':size, \
                        'symbolPen':pg.mkPen(None)}
        #brush = pg.mkBrush(color)
        mask = nodes.is_outside
        
        index = 'inside'
        color = self.param.getValues()[index][0] #č tam bude barva
        self.serint.add_serie(nodes[~mask], z=30,\
                    symbolBrush=color, name=index, **plot_params)
        
        index = 'outside'
        color = self.param.getValues()[index][0] #č tam bude barva
        self.serint.add_serie(nodes[mask], z=30, \
                    symbolBrush=color, name=index, **plot_params)
                    
        # keep the GUI responsive :)
        self.sb_item.app.processEvents()
            
            
            
    def integrate(self):
        #č integrace se stala stateful
        #č a musím jú taky testovat
        if 'ghull' not in self.__dict__:
            ghull = self.get_ghull()
        else:
            ghull = self.ghull
        nsim = self.sb_item.slider.value()
        sample = self.sb_item.sample_box.f_model[:nsim]
        #č teď - saháme do vnitřku naších obliběných tříd
        ghull.hull.sample = sample
        ghull.sample = sample
        
        self.draw_convex_hull(ghull.hull)
        
        
        #☺ Krucinal, kdo ten OrderedDict vymyslel?
        params = self.param.getValues()
        budget = params['budget'][0]
        
        if self.param.getValues()['index'][0]: # replace previous
            self.serint.clear()

        try:
            data = ghull.integrate(budget, callback_all=self.integration_callback)
            ghull_estimation, convex_hull_estimation, global_stats = data
            
            #if hasattr(self.sb_item.sample_box, 'estimations'):
                #self.sb_item.sample_box.estimations.append(data)
                #self.sb_item.estimation_added.emit()
            self.table.setData({**global_stats, "ghull_estimation":ghull_estimation,\
                               "convex_hull_estimation": convex_hull_estimation})
                                
            self.draw_ghull(global_stats['r'], global_stats['R'])
            
            
        except ValueError as e: #BaseException
            msg = "error during estimation "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)  
            self.sb_item.errors.append(e)

        # draw DP
        if ghull.hull.space != 'G':
            size = self.param.getValues()['node (pixel) size'][0] + 3
            plot_params = {'pen':None, 'symbol':'t2', 'symbolSize':size, \
                            'symbolPen':pg.mkPen(None)}
            
            #color = self.param.getValues()[index][0] #č tam bude barva
            self.serint.add_serie(ghull.gint.DP, z=35, index='DP', name='DP', **plot_params)


class ContactWidget(pg.LayoutWidget):
    def __init__(self, w,  parent=None):
        super().__init__(parent)
        self.w = w
        
        w.image_view.mouse_moved.connect(self.on_mouse_moved)
        w.image_view.mouse_clicked.connect(self.on_mouse_dragged)
        w.image_view.mouse_double_clicked.connect(self.on_double_click)
        w.image_view.mouse_right_dragged.connect(self.on_mouse_dragged)
        
        w.space_changed.connect(self.on_space_changed)
        w.slice_changed.connect(self.on_nsim_changed)
        #w.redraw_called.connect(self.hide)
        
        n = w.slider.value() 
        size = n * (n - 1) // 2
        self.condensed_contacts = np.zeros(size, dtype=np.int8)
        self.mask = np.tri(n, k=-1, dtype=bool).T
        self.nsim = n
        
        self.setup()
        self.kwargs = {}
        self.check_contact = self.peacefully_check_contact
        self.got_user_consent = False
        self.setup_CS()
        
    def on_nsim_changed(self, n):
        try:
            del self.qframe
            del self.condensed_qontacts
        except AttributeError:
            pass
        size = n * (n - 1) // 2
        gc.collect()
        self.condensed_contacts.resize(size)
        self.condensed_contacts[:] = 0
        self.mask = np.tri(n, k=-1, dtype=bool).T
        self.nsim = n
        
        self.on_space_changed()
    
    def on_space_changed(self, *args, **kwargs):
        self.setup_CS()
    
    
    def on_mouse_moved(self, x, y):
        if (x == -1) or (x == y):
            self.status_label.setText("")
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        try:
            if self.condensed_qontacts[entry]:
                self.status_label.setText("There is DEFINITELY contact between %d and %d" % (x, y))
            else:
                self.status_label.setText("There is definitely NO contact between %d and %d" % (x, y))
        except AttributeError:
            val = self.condensed_contacts[entry]
            if val > 0:
                self.status_label.setText("There IS contact between %d and %d" % (x, y))
            elif val < 0:
                self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            else:
                self.status_label.setText("")
        
        
    def on_mouse_dragged(self, x, y):
        if (x == -1) or (x == y):
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        val = self.check_contact(entry, i, j)
        self.show()
        
        if val > 0:
            self.status_label.setText("There IS contact between %d and %d" % (x, y))
        else:
            self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            
    def on_double_click(self, x, y):
        if (x == -1) or (x == y):
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        val = self.force_check_contact(entry, i, j)
        self.show()
        
        if val > 0:
            self.status_label.setText("There IS contact between %d and %d" % (x, y))
        else:
            self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            
        
    def discover(self):
        nsim = self.nsim
        if  nsim > 0:
            self.stopbtn.setEnabled(True)
            self.stopbtn.setCheckable(True)
            
            if self.param.getValues()['check_all'][0]:
                for i in range(nsim-1, -1, -1):
                    # keep the GUI responsive :)
                    self.show()
                    self.w.app.processEvents()
                    if self.stopbtn.isChecked():
                        break
                    preentry = nsim * i - ((i + 2) * (i + 1)) // 2
                    for j in range(i+1, nsim):
                        self.check_contact(preentry + j, i, j)
            else:
                failsi = self.w.get_sample_box().failsi
                for i in range(nsim-1, -1, -1):
                    # keep the GUI responsive :)
                    self.show()
                    self.w.app.processEvents()
                    if self.stopbtn.isChecked():
                        break
                    preentry = nsim * i - ((i + 2) * (i + 1)) // 2
                    if failsi[i]: #č první je červený
                        for j in range(i+1, nsim):
                            self.check_contact(preentry + j, i, j)
                    else:
                        for j in range(i+1, nsim):
                            if failsi[j]:
                                self.check_contact(preentry + j, i, j)
            self.stopbtn.setChecked(False)
            self.stopbtn.setEnabled(False)
            self.show()


    def force_check_contact(self, entry, i, j):
        val = -1 + 2 * self.CS.is_couple((i, j), **self.kwargs)
        self.condensed_contacts[entry] = val
        #č není to úplně ideální kód style. Nastavit a zaroveň vrácet :(
        return val 
        
    
    def peacefully_check_contact(self, entry, i, j):
        val = self.condensed_contacts[entry]
        if val == 0:
            return self.force_check_contact(entry, i, j)
        else:
            return val
        
        
    def setup(self):
        self.toolbar = QtWidgets.QToolBar(self)
        
        action = self.toolbar.addAction("discover", self.discover)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Searches for a common facets between Voronoi cells")
        
        action = self.toolbar.addAction("stop", self.stop)
        self.stopbtn = btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setCheckable(True)
        btn.setChecked(False)
        btn.setEnabled(False)
        
        
        action = self.toolbar.addAction("Qhull check", self.check)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Compare with exact Qhull wireframe")
        
        action = self.toolbar.addAction("show", self.show)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("hide", self.hide)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("clear", self.clear)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        self.addWidget(self.toolbar, row=0, col=0)
        
        
        #оӵ остальной (люкет) уллапала
        
        ### Create ParameterTree widget
        self.ptree = pg.parametertree.ParameterTree()
        self._set_param()
        self.ptree.setParameters(self.param, showTop=False)
        
        self.addWidget(self.ptree, row=1, col=0)
        
        self.status_label = QtWidgets.QLabel()
        #self.status_label.setText()
        self.addWidget(self.status_label, row=2, col=0)
        
    def stop(self):
        self.stopbtn.setChecked(True)
        
    def qframe_callback(self):
        # keep the GUI responsive :)
        self.w.app.processEvents()
        
        self.dlg += 1
        return self.dlg.wasCanceled()
  
    
            
    @staticmethod
    def get_user_consent():
        return QtWidgets.QMessageBox.question(None, 
            "Performing triangulation in high dimensions may take a while...",
            "Would you like to start Qhull anyway?"
            ) == QtWidgets.QMessageBox.Yes
        
        
    def get_qhull(self):
        try:
            return True, self.qframe
        except AttributeError:
            sample_box = self.w.get_sample_box()
            if (sample_box.nvar < 10) or self.got_user_consent or self.get_user_consent():
                self.got_user_consent = True
                sample_space = getattr(sample_box, self.w.space)
                self.qframe = wireframe.Qframe(sample_space)
                return True, self.qframe
            else:
                return False, None
        
        
    def check(self):
        if self.get_qhull()[0]:
            with pg.ProgressDialog("Going over all the simplices..", 0, 
                                    self.qframe.nsimplex, cancelText='Stop', 
                                    busyCursor=True) as dlg:
                self.dlg = dlg
                self.qframe.generate_wireframe(self.qframe_callback)
                if not self.dlg.wasCanceled():
                    self.condensed_qontacts = self.qframe.condensed_contacts
                    self.show()
            
        
        
    def clear(self):
        self.condensed_contacts[:] = 0
        try:
            del self.qframe
            del self.condensed_qontacts
        except AttributeError:
            pass
            
        self.w.image_view.blue[self.mask] = -1
        self.w.image_view.update()
    
    def show(self):
        try:
            self.w.image_view.blue[self.mask] = self.condensed_qontacts * 2 - 1
            self.w.image_view.blue[self.mask] -= self.condensed_contacts
            self.w.image_view.blue[self.mask] /= 2
        except AttributeError:
            self.w.image_view.blue[self.mask] = self.condensed_contacts
        self.w.image_view.update(autoRange=False)
    
    def hide(self):
        self.w.image_view.blue[self.mask] = -1
        self.w.image_view.update()

        
    def _set_param(self):
        params = list()
        #č check_all ovlivňuje discover(). discover ho i kontroluje
        params.append({'name': 'check_all', 'title': 'all contacts', 
                        'type': 'bool', 'value': True,
                        'tip': 'allows to skip "green" contacts'})
        params.append({'name': 'force_update', 'title': 'reevaluate', 
                        'type': 'bool', 'value': False,
                        'tip': "always reevaluate even discovered contacts"})
        params.append({'name': 'method', 'type': 'list', 'value': 'ConvexLinear', 
                     'values': ['DirectContact', 'Gabriel', 'ConvexSolver', 'ConvexLinear', 'LinearSolver', 'LocalizedHull', 'Qhull'],
                     'title': "method"})
        
        params.append({'name': 'solver', 'title': "solver", 'type': 'list', 'value': 'convex_sprite',
                    'values': ['convex_solver', 'convex_spline', 'convex_sort', 'convex_sprite', 'convex_slice'],  
                    'tip': "Versions from the earliest (the worst) to the latest (the best). Names have no any special meaning."})
                     
        params.append({'name': 'tries_to_fix', 'type': 'int', 
                    'limits': (0, float('inf')), 'value': 1, 'default': 1, 
                    'title': "additional tries", 
                    'tip': "Number of additional (over required ndim) tries to rotate hyperplane and find the contact."})   
            
        params.append({'name': 'tol', 'type': 'float', 
                    'title': "tolerance", 
                    'limits': (0, float('inf')), 'value': 1e-5, 'default': 1e-7})   
                    
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        #č chlapi tomu svému stromu dali zabrat
        #č udělali i signaly
        p = self.param.child('force_update')
        p.sigValueChanged.connect(self.check_method_changed)
        
        p = self.param.child('solver')
        p.sigValueChanged.connect(self.cs_solver_changed)
        
        p = self.param.child('method')
        p.sigValueChanged.connect(self.method_changed)
        
        
        p = self.param.child('tries_to_fix')
        p.sigValueChanged.connect(self.ttf_changed)
        
        p = self.param.child('tol')
        p.sigValueChanged.connect(self.tol_changed)
        
    
    def ttf_changed(self, parameter):
        self.kwargs['tries_to_fix'] = parameter.value()
        
    def tol_changed(self, parameter):
        self.kwargs['tol'] = parameter.value()
        
        
    def check_method_changed(self, parameter):
        if parameter.value():
            self.check_contact = self.force_check_contact
        else:
            self.check_contact = self.peacefully_check_contact
        
    def cs_solver_changed(self, parameter):
        #č ale.. lezeme do vnítřku cizí třídy..
        solver = getattr(wireframe, parameter.value())
        self.CS.convex_solver = solver
        
    def method_changed(self, parameter):
        self.setup_CS()
        method = parameter.value()
        if method in ('ConvexSolver', 'ConvexLinear'):
            self.param.child('solver').show()
            self.param.child('tries_to_fix').show()
            self.param.child('tol').show()
        elif method in ('LocalizedHull', 'LinearSolver'):
            self.param.child('solver').hide()
            self.param.child('tries_to_fix').hide()
            self.param.child('tol').show()
        else:
            self.param.child('solver').hide()
            self.param.child('tries_to_fix').hide()
            self.param.child('tol').hide()
        
#        if method == 'LinearSolver':
#           self.param.child('force_update').setReadonly()
#           self.param.child('force_update').setValue(False)
#        else:
#           self.param.child('force_update').setReadonly(False)
        
        
    def setup_CS(self):
        sample_box = self.w.get_sample_box()
        sample_space = getattr(sample_box, self.w.space)
        
        #č kvůli Qhull se může stát, že budeme muset zachovat starý CS
        #č proto opatrně mažeme kwargs no a tak.
        method = self.param.child('method').value()
        if method == 'Qhull':
            status, qframe = self.get_qhull()
            if status:
                self.CS = qframe
                self.kwargs.clear()
            else:
                self.param.child('method').setToDefault()
        elif method == 'DirectContact':
            self.CS = wireframe.DirectContact(sample_space)
            self.kwargs.clear()
        elif method == 'Gabriel':
            self.CS = wireframe.Gabriel(sample_space)
            self.kwargs.clear()
        elif method == 'ConvexSolver':
            solver = getattr(wireframe, self.param.child('solver').value())
            self.CS = wireframe.ConvexSolver(sample_space, convex_solver=solver)
            #self.kwargs.clear() #č aby se to neztratilo až bude 100500 metodů
            self.kwargs['tries_to_fix'] = self.param.getValues()['tries_to_fix'][0]
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'ConvexLinear':
            solver = getattr(wireframe, self.param.child('solver').value())
            self.CS = wireframe.ConvexLinear(sample_space, convex_solver=solver)
            #self.kwargs.clear() #č aby se to neztratilo až bude 100500 metodů
            self.kwargs['tries_to_fix'] = self.param.getValues()['tries_to_fix'][0]
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'LinearSolver':
            self.CS = wireframe.LinearSolver(sample_space)
            self.kwargs.clear()
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'LocalizedHull':
            self.CS = wireframe.LocalizedHull(sample_space)
            self.kwargs.clear()
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        else:
            raise ValueError("ContactWidget: unknown method")
            
        
        




#!/usr/bin/env python
# coding: utf-8


import numpy as np
from scipy.spatial import distance
from scipy.spatial import KDTree

import quadpy

from .. import shell
from .. import simplex as sx
from .. import convex_hull as khull
from ..reader import Store
from ..candynodes import CandyNodes
from .. import sball

from ._exploration import _Exploration

from collections import namedtuple
from sortedcollections import ValueSortedDict





def get_entropy(pf):
    entropy = -pf * np.log(pf) - (1 - pf) * np.log(1 - pf)
    if entropy > 0:
        return entropy
    else:
        return 0

max_entropy = get_entropy(0.5)

    
    
TriEstimation = namedtuple('TriEstimation', (
    *shell.ShellStats._fields,
    *shell.ShellEstimation._fields,
    'success_points',
    'failure_points',
    'success',
    *sx.FastCubatureEstimation._fields[2:],
    'max_potential'
    ))


class CircumTri(_Exploration):
    
    #č míží nám sampling_space: Ghull umí vzorkovat outside pouze v G prostoru
    #č quadpy umístí integráční bodíky v prostoru triangulace.
    def __init__(bx, sample_box, scheme, q=10, #circumcenters_only=True, 
                  store_candidates_metainformation=False, 
                  potential_mode=2, weighted_entropy=True, shell_budget=1000):
        
        bx.sample_box = sample_box
        bx.scheme = scheme
        
        bx.tri_space = 'G'
            
        
        bx.shell_budget = shell_budget
        bx.potential_mode = potential_mode
        bx.weighted_entropy = weighted_entropy
        bx.store_candidates_metainformation = store_candidates_metainformation
        #bx.circumcenters_only = circumcenters_only
        bx.q = q
        
        
        bx.direct_plan = quadpy.un.mysovskikh_1(bx.nvar).points
        bx.sball = sball.Sball(bx.nvar)
        
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_tri", TriEstimation)
        else:
            bx.box_estimations = []
            
        bx.CC = sx.CircumCenter(sample_box.nvar)
            
        # kind of interface to CandidatesWidget
        bx.candidates_index = {}
        bx.potential_index = ValueSortedDict()
        bx.rating_index = ValueSortedDict()
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 'scheme':bx.scheme.name,
                 'potential_mode':bx.potential_mode,\
                 'shell_budget':bx.shell_budget,
                 #'circumcenters_only':bx.circumcenters_only, 
                 'q':bx.q}
    
    def __repr__(bx):
        return "%s(**%s)"%(bx.__class__.__name__,  repr(bx.init_parameters()))
        
    def __str__(bx):
        return "%s(%s)"%(bx.__class__.__name__,  str(bx.init_parameters()))
        
        
    def refine(bx):
        if len(bx.candidates_index):
            # potential mode:
            # 0 - never use, always use rating
            # 1 - always select by potential
            # n - select by potential every n-th node
            if bx.potential_mode < 1:
                # return node with the greatest rating
                key, value = bx.rating_index.peekitem()
            elif bx.nsim % bx.potential_mode: 
                # return node with the greatest rating
                key, value = bx.rating_index.peekitem()
            else:
                # return node with the greatest potential
                key, value = bx.potential_index.peekitem()
            return bx.candidates_index[key]
        else:
            bx._logger(msg='refine called, but triangulation does not exist yet. Fallback to random sample')
            return bx.f_model(1)
    
    
    def regen(bx):
        """
        regen() recreates data structures of the box. 
        It shouldn't be called without reason, changed distribution, settings or so.
        """
        
        #оӵ шайтан регенираци лэзьиз
        bx._logger(msg='regeneration started')
        
        bx.candidates_index.clear()
        bx.potential_index.clear()
        bx.rating_index.clear()
        
        bx._regen_outside()
        bx._regen_inside()
        
        if bx.nsim > 0:
            bx.get_pf_estimation() #č updates p_mixed
            bx.update_exploration_ratio()
        else:
            bx.to_explore = 1
            bx.to_refine = 0
        
        
        bx._nsim = bx.nsim
                 
    
    
    
    
    def __len__(bx):
        return bx.sample_box.nsim
    
    def __getitem__(bx, slice):
        #č stačí vratit sample_box
        return bx.sample_box[slice]
    
    def __getattr__(dx, attr):
        if attr == 'dicebox':
            return dx
            
        #č branime sa rekurzii
        # defend against recursion
        #оӵ рекурсилы пезьдэт!
        if attr == 'sample_box':
            raise AttributeError
                
        #ё По всем вопросам обращайтесь 
        #ё на нашу горячую линию    
        else:
            return getattr(dx.sample_box, attr)
    
    
    # The DiceBox Observer 
    def _logger(self, *args, msg="", indent=0, **kwargs):
        if not kwargs:
            kwargs = "" #č ať se nám prázdné závorky nezobrazujou
        print(self.__class__.__name__ + ":", msg, *args, kwargs) 
        
        
        # inspired by Qt
    def connect(self, slot): self._logger = slot
    def disconnect(self): del(self._logger)
    
        
    # přidávání vzorků musí bejt explicitní!
    def add_sample(bx, input_sample):
        bx.sample_box.add_sample(input_sample)
        bx.increment(bx.sample_box[bx._nsim:])
        bx._nsim = bx.nsim
    
    
    def update_exploration_ratio(bx):
        bx.to_explore += np.sqrt(bx.shell_estimation.outside)
        bx.to_refine += np.sqrt(bx.p_mixed)
    
    
    
    #č teď je to kolbek, který volá Triangulation
    def _on_mixed_added(bx, simplex, indices, vertices_model, nodes, _vol, fr, wfr, _mfpdf):
        
        circum_center = bx.CC.get_circumcenter(vertices_model)
        r = distance.euclidean(circum_center, vertices_model[0])
        circum_node = bx.f_model.new_sample(circum_center, space=bx.tri_space)
        #č můžeme nechat numpy pole z jednoho prvku. Můžeme zredukovat na float
        #circum_pdf = circum_node.pdf(bx.tri_space)
        circum_pdf = float(circum_node.pdf(bx.tri_space))
        circum_potential = r * circum_pdf**(1/bx.nvar)
        
        failsi = bx.Tri.failsi[indices]
        
        
        tree = KDTree(vertices_model, compact_nodes=False, balanced_tree=False)
        
        #č konkretně tato třída je pevně napojena na G prostor
        #č ale bacha, kbyby se to změnilo...
        dd, ii = tree.query(nodes.G, k=2)
        
        failsi_pair = failsi[ii]
        mask = np.logical_xor(failsi_pair[:,0], failsi_pair[:,1])
        #mask = np.sum(failsi[ii], axis=1) == 1
        if np.any(mask):
            nodes = nodes[mask]
            
            dr = distance.cdist(nodes.G, [circum_center]).reshape(-1)
            nodes_pdf = nodes.pdf('G')
            node_potentials = (r - dr) * nodes_pdf**(1/bx.nvar)
            
            
            max_node = np.nanargmax(node_potentials)
            max_node_potential = node_potentials[max_node]
        else:
            max_node_potential = -np.inf
        
        
        
        if max_node_potential > circum_potential:
            result_node = nodes[max_node]
            result_potential = max_node_potential
            
            d1, d2 = dd[mask][max_node]
            if bx.weighted_entropy:
                ii = ii[mask][max_node]
                pdf1, pdf2 = bx.Tri.PDF[indices][ii]
                entropy = get_entropy(pdf1 * d2 / (pdf1 * d2 + pdf2 * d1))
            else:
                entropy = get_entropy(d1 / (d1 + d2))
            result_rating = max_node_potential * entropy
        else:
            result_node = circum_node
            result_potential = circum_potential
            # circum rating
            if bx.weighted_entropy:
                entropy = get_entropy(wfr)
            else:
                entropy = get_entropy(fr)
            result_rating = circum_potential * entropy
        
        
        #č nodes příjdou zabalené do CandyNodes. Ty mají .attrs a .kwargs
        if bx.store_candidates_metainformation:
            result_node = CandyNodes(result_node)
            result_node.potential = result_potential
            result_node.rating = result_rating
            
        bx.candidates_index[simplex] = result_node
        bx.potential_index[simplex] = result_potential
        bx.rating_index[simplex] = result_rating
            
        
            

            
    # callback
    #č sx.on_delete_simplex(indices=indices)
    def _invalidate_simplex(bx, simplex):
        bx.potential_index.pop(simplex, None)
        bx.rating_index.pop(simplex, None)
        bx.candidates_index.pop(simplex, None)
    
    
    
        

                
    def _regen_outside(bx):
        bx.convex_hull = khull.QHull(bx.f_model, space='G', 
                                    incremental=True, auto_update=False)
        bx.ghull = shell.GaussianAnnulus(bx.convex_hull)
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        if bx.nsim > 0:
            bx.estimate_outside()
        
            
    def _regen_inside(bx):
        failsi = bx.failsi
        # incremental triangulation require one more point
        if (bx.nsim > bx.nvar + 1) and np.any(failsi) and not np.all(failsi):
            #bx._logger(msg="triangulation started")
            bx.__regen_inside()
        else:
            #č jíž není nutný
            bx._logger(msg="triangulation skipped")
            
    def __regen_inside(bx):
        try:
            bx.Tri = sx.CubatureIntegration(bx.samplebox, bx.scheme,\
                     tri_space=bx.tri_space, incremental=True,\
                    on_mixed_added=bx._on_mixed_added,\
                    on_delete_simplex=bx._invalidate_simplex)
              
            bx.Tri.integrate() # nic nevrácí, všecko je přes kolbeky
            #č tri - Deloneho triangulace
            bx.tri = bx.Tri.tri #č všichní tam očekávajou QHull
            bx._logger(msg="triangulation has been created")
            
        except BaseException as e:
            #č chcu zachytit spadnuti QHull na začatku, 
            #č kdy ještě není dostatek teček.
            #č Jinak je třeba nechat QHull spadnout
            if bx.nsim > 2*bx.nvar + 3: 
                #č no to teda ne!
                raise
            else: 
                #č lze přípustit chybu triangulace    
                bx._logger(msg='triangulation failed')
    
    
        
    def increment(bx, input_sample):
        #č tri - Deloneho triangulace
        if "tri" in dir(bx):
            bx.Tri.update()
        else:
            bx._regen_inside()
            
            
        if np.any(bx.convex_hull.is_outside(input_sample)):
            bx.to_explore = 0
            bx.convex_hull.update()
            bx.estimate_outside()
        else:
            bx.to_refine = 0
        
        bx.box_estimations.append(bx.get_pf_estimation())
        bx.update_exploration_ratio()
    
                
                    
    
    
    
    def _rate_outside_nodes(bx, outside_nodes, d):
        #č sice získáme filtrovaný outside, 
        node_potentials = outside_nodes.pdf(bx.tri_space)**(1 / bx.nvar) * d
        
        max_node = np.nanargmax(node_potentials)
        max_node_potential = node_potentials[max_node]
        
        
        if max_node_potential > bx.potential_index[-1]:
            node_rating = 0 #max_node_potential * max_entropy
            if bx.store_candidates_metainformation:
                outside_node = CandyNodes(outside_nodes[max_node])
                outside_node.potential = max_node_potential
                outside_node.rating = node_rating
            else:
                outside_node = outside_nodes[max_node]
            
            bx.candidates_index[-1] = outside_node
            bx.potential_index.pop(-1)
            bx.potential_index[-1] = max_node_potential
            bx.rating_index.pop(-1, None)
            bx.rating_index[-1] = node_rating
        
        
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        #č Máme 2 úkoly: 
        #č 1. Získat odhady a uložit je, abychom nemuseli opakovaně integrovat,
        #č    dokud se neobjeví nějaký nový vzorek zvenku.
        #č 2. Získat kandidaty.
        #č    a. z mezíkruží (-12)
        
        #č prace s tečkami v mezikruži se změnila
        #č teď tečky dostávám přes kolbek po částech
        #č a není předem známo, kolik těch částí bude.
        bx.candidates_index.pop(-1, None)
        #č těch kastomných slovníků se bojím...
        bx.potential_index.pop(-1, None)
        #bx.rating_index.pop(-1, None)
        bx.potential_index[-1] = 0 #č pro callback
        
        bx.shell_stats = bx.ghull.get_shell_estimation()
        
        # get candidates!
        #č explicitně (pokažde) počtem teček zadavám přesnost integrace
        #č takže změny bx.shell_budget budou při dálším spuštění aplikovány
        bx.shell_estimation = bx.ghull.integrate(bx.shell_budget, \
                                callback_outside=bx._rate_outside_nodes) 
        
    
    
    def get_pf_estimation(bx):
        failsi = bx.failsi
        
        success_points = len(failsi[~failsi])
        failure_points = len(failsi[failsi])
        
        inside = bx.shell_estimation.inside
        
        __key, max_potential = bx.potential_index.peekitem()
        
        if 'tri' in dir(bx):
            tri_estimation = bx.Tri.get_pf_estimation()
            
            bx.p_mixed = tri_estimation.mix
            
            success = inside - tri_estimation.failure - tri_estimation.mix
            
            return TriEstimation(bx.nsim,  #č musíme sami lepit nové etikety, neboť
                                *bx.shell_stats[1:], #č Ghull spouštíme sporadicky
                                *bx.shell_estimation,
                                success_points,
                                failure_points,
                                success,
                                *tri_estimation[2:],
                                max_potential
                                )
            
        
        #оӵ триангуляци ӧвӧл, иськем...
        
        #č může se stát, že první dvě tečky už hned májí různé barvy,
        #č ale žádnej simplex ještě nemáme.
        #č takže celou skříňku prostě bereme jako simplex
        event, event_id, fr, wfr = sx.get_simplex_event(bx, weighting_space=bx.tri_space)
        # -1=outside, 0=success, 1=failure, 2=mix
        tri_estimation = {0:0, 1:0, 2:0}
        tri_estimation[event_id] = inside
        
        success = tri_estimation[0]
        failure = tri_estimation[1]
        bx.p_mixed = mixed = tri_estimation[2]
        
        return TriEstimation(bx.nsim,  #č musíme sami lepit nové etikety, neboť
                             *bx.shell_stats[1:], #č Ghull spouštíme sporadicky
                             *bx.shell_estimation,
                             success_points,
                             failure_points,
                             success,
                             
                             failure,
                             mixed,
                             inside * fr,
                             inside * wfr, 
                             inside * wfr,
                             0, # nsimplex
                             #sx.tn_scheme.name,
                             bx.scheme.points.shape[1],
                             0, #sx.newly_invalidated,
                             0, #sx.newly_estimated,
                             0, #len(sx.failure_simplices),
                             0, #len(sx.mixed_simplices),
                             0, #len(sx.tri.coplanar),
                             
                             max_potential
                             )
        
        



#!/usr/bin/env python
# coding: utf-8

# Axes level functions (for Matplotlib)
# Tier3 module => suppose dirty unsupported code

#č Založme nový modul pro funkce (procedury),
#č které sice stejně jako mart pracují na urovni axes,
#č ale na rozdil od atomických funkcí mart modulu
#č udělaj z prazdných os hotový obrazek.


from . import mart
from . import mart3d

# it is mostly for qt_plot, it offers availiable options to user
__all__ = ['tri_plot', 'plot3D', 'plot_points_3D',
         'candidates_drawing', 'plane_under_density', 
         'qhull_under_density', 'dhull_under_density']

def plot3D(ax3d):
    mart3d.scatter_points(ax3d)
    
    ax3d.set_xlabel("$x_1$")
    ax3d.set_ylabel("$x_2$")
    ax3d.set_zlabel("$x_3$")


def plot_points_3D(ax3d):
    mart3d.scatter_points(ax3d, markers=('.', '.', '.', '.'), alpha=0.5, rasterized=True)
    
    ax3d.set_xlabel("$x_1$")
    ax3d.set_ylabel("$x_2$")
    ax3d.set_zlabel("$x_3$")


#č ten [plot] zásadně vytvaří své struktury, nepouzívá oné ze skříňky,
#č protože já vím, že v těch obrázcích, ve kterých chcu ho použit,
#č můde být třeba použit řez a skříňka tedy potřebné struktury může nemít
def tri_plot(ax3d, tri_space=None, lw=0.3, nrid=200):
    from .. import simplex as six
    if tri_space is None:
        tri_space = ax3d.space
    
    Tri = six.JustCubatureTriangulation(ax3d.sample_box, tn_scheme=None, \
                                    tri_space=tri_space, issi=None,\
                                    weighting_space=None, \
                                    incremental=False,\
                                    on_add_simplex=None,\
                                    on_delete_simplex=None)
    
    if tri_space == ax3d.space:
        mart3d.tri_plot(ax3d, Tri=Tri, ns=2, color="#74747445", lw=lw/1.4, zorder=100, rasterized=True)
    else:
        mart3d.tri_plot(ax3d, Tri=Tri, ns=nrid, color="#74747445", lw=lw/1.4, zorder=100, rasterized=True)
    
    mart3d.scatter_points(ax3d)
    
    ax3d.set_xlabel("$x_1$")
    ax3d.set_ylabel("$x_2$")
    ax3d.set_zlabel("$x_3$")
    

    
def candidates_drawing(ax):
    try:
        mart.triplot(ax, color="grey", linewidth=0.4)
    except:
        pass
    mart.plot_boundaries(ax, linewidth=0.7)
    
    #č ax.scatter posílá parameter cmap Collections třídě.
    #č Třída mimo jiného dědí cm.ScalarMappable,
    #č která inicializaci deleguje funkci cm.get_cmap() ve (svém) modulu cm.
    # cmap='viridis_r' #cmap='plasma',
    mart.scatter_candidates(ax, s=5, marker='.', cmap='plasma_r',\
     alpha=None, linewidths=None, edgecolors=None, plotnonfinite=False)
    mart.plot_the_best_candidate(ax, "^", color='#3D0D5B')
    
    mart.scatter_points(ax)    
    
    
    
def plane_under_density(ax3d, lim=3, pudorys=-0.2, jemnost=30):
    mart.scatter_points(ax3d, zs=pudorys, s=3)#, zorder=300000)
    #ax3d.set_zbound()
    #ax3d.set_frame_on(False)
    #ax3d.grid(False)
    
    ax3d.margins(0)
    ax3d.set_xlim(-lim, lim)
    ax3d.set_ylim(-lim, lim)
    #ax3d.set_xlabel('$x_1$') #ošklivý
    #ax3d.set_ylabel('$x_2$') #ošklivý
    #ax3d.set_axis_off() # vypne aj dolní panev
    #ax3d.set_zlim(pudorys, 5)
    ax3d.set_xticklabels([])
    ax3d.set_yticklabels([])
    
    mart3d.density_surface(ax3d, alpha=0.8, rcount=50, edgecolor='black', linewidth=0.3, color='#4472C4')
    
    
def qhull_under_density(ax3d, lim=3, pudorys=-0.2, jemnost=30, hezkymodře=(85/255, 70/255, 1, 1)):
    import numpy as np
    #from matplotlib import colors
    from .. import convex_hull as khull

    qhull = khull.QHull(ax3d.sample_box, space=ax3d.space)
    
    mart.scatter_points(ax3d, zs=pudorys, s=3)#, zorder=300000)
    #ax3d.set_zbound()
    #ax3d.set_frame_on(False)
    #ax3d.grid(False)
    
    ax3d.margins(0)
    ax3d.set_xlim(-lim, lim)
    ax3d.set_ylim(-lim, lim)
    #ax3d.set_xlabel('$x_1$') #ošklivý
    #ax3d.set_ylabel('$x_2$') #ošklivý
    #ax3d.set_axis_off() # vypne aj dolní panev
    #ax3d.set_zlim(pudorys, 5)
    ax3d.set_xticklabels([])
    ax3d.set_yticklabels([])
    
    mart3d.density_colored_surface(ax3d, qhull, alpha=0.8, rcount=50, edgecolor='black', linewidth=0.3, colors=('#B9E40E', '#808080'))
    
    list_of_lines2D = mart.qhull_plot(ax3d, qhull, zs=pudorys, linewidth=0.5, color=hezkymodře)
    for line2d in list_of_lines2D:
        mart3d.wall(ax3d, line2d, zmin=0, zmax=0.03, facecolor=hezkymodře, edgecolor=hezkymodře, linewidth=0.0, alpha=1)
    


def dhull_under_density(ax3d, ndir=10, lim=3, pudorys=-0.2, jemnost=30, hezkymodře=(85/255, 70/255, 1, 1)):
    import numpy as np
    #from matplotlib import colors
    from .. import convex_hull as khull
    from .. import sball
    
    direct_plan = sball.get_random_directions(ndir, 2)
    dhull = khull.DirectHull(ax3d.sample_box, direct_plan, space=ax3d.space)
    
    mart.scatter_points(ax3d, zs=pudorys, s=3)#, zorder=300000)
    #ax3d.set_zbound()
    #ax3d.set_frame_on(False)
    #ax3d.grid(False)
    
    ax3d.margins(0)
    ax3d.set_xlim(-lim, lim)
    ax3d.set_ylim(-lim, lim)
    #ax3d.set_xlabel('$x_1$') #ošklivý
    #ax3d.set_ylabel('$x_2$') #ošklivý
    #ax3d.set_axis_off() # vypne aj dolní panev
    #ax3d.set_zlim(pudorys, 5)
    ax3d.set_xticklabels([])
    ax3d.set_yticklabels([])
    
    mart3d.density_colored_surface(ax3d, dhull, alpha=0.8, rcount=50, edgecolor='black', linewidth=0.3, colors=('#B9E40E', '#808080'))
    
    list_of_lines2D = mart.dhull_plot(ax3d, dhull, linewidth=0.5, color=hezkymodře)
    



#!/usr/bin/env python
# coding: utf-8

# Monkey patch module
# Just import it to make MPL unpredictable

# taken from StackOverflow
# https://stackoverflow.com/questions/16488182/removing-axes-margins-in-3d-plot
# © HYRY, 2013
# license: CC BY-SA 3.0
###patch start###
from mpl_toolkits.mplot3d.axis3d import Axis
if not hasattr(Axis, "_get_coord_info_old"):
    def _get_coord_info_new(self, renderer):
        mins, maxs, centers, deltas, tc, highs = self._get_coord_info_old(renderer)
        mins += deltas / 4
        maxs -= deltas / 4
        return mins, maxs, centers, deltas, tc, highs
    Axis._get_coord_info_old = Axis._get_coord_info  
    Axis._get_coord_info = _get_coord_info_new
###patch end###



#!/usr/bin/env python
# coding: utf-8

import sys
from . import qt_testcases
from . import qt_dicebox
from .. import reader
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets


def select_test_case():
    return qt_testcases.SelectTestCaseWidget().box


def read_box(wt):
    """
    Funkce read_box() buď vytvoří Reader object
    kterým nahradí wt.sample_box,
    nebo - je-li uživatel odmitné zvolit nějaký soubor
    ponechá původní samplebox
    """ 
    app = pg.mkQApp()
    csv_filter = "CSV files (*.csv)"
    filename, __filter = pg.FileDialog.getSaveFileName(
            parent=None, 
            caption="open/store",
            directory="store/%s_%sD.csv" % (wt.gm_signature, wt.nvar),
            #filter=csv_filter,
            filter=csv_filter + ";;All files(*)", 
            initialFilter=csv_filter,
            options=pg.FileDialog.Option.DontConfirmOverwrite
            )
    if filename:
        #č Reader vždy dopíše '.csv' na konci souboru
        #č nechcu to teď měnit
        if filename[-4:]=='.csv':
            filename = filename[:-4]
        #č Reader sám zavolá SampleBox 
        wt.sample_box = reader.Reader(filename, wt.samplebox.sampled_plan)
        
def setup_dicebox(wt):
    qt_dicebox.SetupDiceBoxWidget(wt)
            

def show_box(box):
    #č pokud jsme ve 3D a uživatel vysloveně chce,
    #č otevřeme mu 3D okinko
    #č jinak import qt_plot a jedeme
    
    if box.nvar == 3:
        answer = QtWidgets.QMessageBox.question(None, "Show 3D case", "Would you like to start 3D view window?")
        if answer == QtWidgets.QMessageBox.Yes:
            from . import gl_plot
            gl_plot.QtGuiPlot3D(box)
            
            #č já nevím, co uživatel bude chtit,
            #č ale asi bych ho neotravoval 2D pohledem
            #č po uzavření 3D okna
            sys.exit()
    
    #č defaultní vetev
    from . import qt_plot
    qt_plot.QtGuiPlot2D(box)




#!/usr/bin/env python
# coding: utf-8

import numpy as np
from scipy import interpolate

from . import sball # for Isocurves
from . import IS_stat

def pf_entropy(pf):
    return -pf*np.log(pf) - (1-pf)*np.log(1-pf)


class RBF_surrogate:
    def __init__(self, sample_box, space='R'):
        self.sample_box = sample_box
        self.space = space
        self._nsim = -100500
        self.update()
    
    def __call__(self, *args):
        self.update()
        self.rbf(*args)
    
    def update(self):
        if self.sample_box.nsim > self._nsim:
            sample_space = getattr(self.sample_box, self.space)
            self.rbf = interpolate.Rbf(*sample_space.T, self.sample_box.g_values, function='gaussian')
            self._nsim = self.sample_box.nsim

    def get_pf_estimation(self, nis=100000):
        self.update()
        nodes = IS_stat.IS_norm(self.sample_box.f_model, mean=0, std=2.5, sampling_space='G', nis=nis, design=None)
        gi = self.rbf(*getattr(nodes, self.space).T)
        return np.sum(nodes.w[gi<0])/nis



def get_isodistances(f_model, r, nrod=200):
    phi = np.linspace(0, 6.283185307, nrod, endpoint=True)
    cos_phi = np.cos(phi)
    sin_phi = np.sin(phi)
    
    sample_G = np.array((cos_phi, sin_phi)).T * r
    return f_model.new_sample(sample_G, space='G', extend=True)


def isolevels_2d(pdf, weighting_pdf_const, r_levels, from_top=None):
    """
    weighting_pdf_const = 1 / (xmax - xmin) / (ymax - ymin)
    """
    s_ball = sball.Sball(2) # nvar=2
    p_levels = []
    for r in r_levels:
        p_levels.append(1 - s_ball.get_pf(r))
    
    return isolevels(pdf, weighting_pdf_const, p_levels, from_top)




def isolevels(pdf, weighting_pdf_const, p_levels, from_top=None):
    """
    weighting_pdf_const = 1 / (xmax - xmin) / (ymax - ymin)
    """
    #č třeba P prostor doopravdy zlobí, takže zkusím nějak tak
    if from_top is None:
        weights = pdf / weighting_pdf_const
        p_all = np.sum(weights) / len(pdf)
        #č prečo víme, že celková pravděpodobnost může bejt nekoněčně velká
        if p_all <= 1:
            from_top = True
        else:
            from_top = False
            
    max_pdf = np.max(pdf)
    pdf_levels = []
    if from_top:
        # descending
        sorted_pdf = np.flip(np.sort(pdf))
        p_cumsum = np.cumsum(sorted_pdf) / weighting_pdf_const / len(pdf)
        for p in p_levels:
            # little bit tricky, didn't find numpy method for this
            mask = p_cumsum <= p
            level_down_bound = np.max(sorted_pdf[~mask], initial=0)
            level_up_bound = np.min(sorted_pdf[mask], initial=max_pdf)
            pdf_levels.append((level_down_bound + level_up_bound) / 2)
            
    else: # from bottom
        sorted_pdf = np.sort(pdf)
        p_cumsum = np.cumsum(sorted_pdf) / weighting_pdf_const / len(pdf)
        for p in p_levels:
            # little bit tricky, didn't find numpy method for this
            mask = p_cumsum <= 1-p
            level_down_bound = np.max(sorted_pdf[mask], initial=0)
            level_up_bound = np.min(sorted_pdf[~mask], initial=max_pdf)
            pdf_levels.append((level_down_bound + level_up_bound) / 2)

    return pdf_levels
